import plotly.graph_objects as go
from plotly.subplots import make_subplots


class Plotter:

    def __init__(self, df, **kwargs):
        self.df = df
        for key, value in kwargs.items():
            setattr(self, key, value)

    def raw_data_chart(self):
        fig = make_subplots(specs=[[{'secondary_y': True}]])
        # Add traces
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Inlet_Pressure'] * 1e-6, name='Eingangsdruck',
                       line_color='blue'),
            secondary_y=False)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Outlet_Pressure'] * 1e-6, name='Ausgangsdruck',
                       line_color='blue'),
            secondary_y=False)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Confining_Pressure'] * 1e-6, name='Manteldruck',
                       line_color='black', visible='legendonly'),
            secondary_y=False)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Temperature'] - 273.15, name='Temperatur',
                       line_color='red', visible='legendonly'),
            secondary_y=True)
        fig.add_vline(
            x=self.df.iloc[self.start]['Duration'], line_dash='dot')
        fig.add_vline(
            x=self.df.iloc[self.stop - 1]['Duration'], line_dash='dot')

        fig.update_layout(
            title=f'<b>{self.name}</b>',
            xaxis=dict(showexponent='all', exponentformat='power'))
        fig.update_xaxes(title_text='Messdauer in s', type='log')
        fig.update_yaxes(title_text='Druck in MPa', secondary_y=False)
        fig.update_yaxes(title_text='Temperatur in °C', secondary_y=True, showgrid=False)
        fig.show()

    def result_chart(self):
        self.df['Diff_Inlet'] = (abs(self.df['Inlet_Pressure'] - self.df['Inlet_Pressure_Cal'])
                                 / self.df['Inlet_Pressure'])
        self.df['Diff_Outlet'] = (abs(self.df['Outlet_Pressure'] - self.df['Outlet_Pressure_Cal'])
                                  / self.df['Outlet_Pressure'])

        fig = make_subplots(specs=[[{'secondary_y': True}]])
        # Add traces
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Inlet_Pressure'] * 1e-6, name='Messwert Eingang',
                       line_color='blue'))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Outlet_Pressure'] * 1e-6, name='Messwert Ausgang',
                       line_color='blue'))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Confining_Pressure'] * 1e-6, name='Manteldruck',
                      line_color='black', visible='legendonly'))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Temperature'] - 273.15, name='Temperatur',
                       line_color='red', visible='legendonly'),
            secondary_y=True)

        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Inlet_Pressure_Cal'] * 1e-6, name='Rechenwert Eingang',
                       line_color='red', mode='markers', marker=dict(size=3)))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Outlet_Pressure_Cal'] * 1e-6, name='Rechenwert Ausgang',
                       line_color='red', mode='markers', marker=dict(size=3)))

        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Diff_Inlet'], name='Abweichung Eingang',
                       line_color='green', fill='tozeroy'),
            secondary_y=True)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Diff_Outlet'], name='Abweichung Ausgang',
                       line_color='yellow', fill='tozeroy'),
            secondary_y=True)

        fig.update_layout(
            xaxis=dict(showexponent='all', exponentformat='power'))
        fig.update_xaxes(title_text='Messdauer in s', type='log')
        fig.update_yaxes(title_text='Druck in MPa', secondary_y=False)
        fig.update_yaxes(secondary_y=True, showgrid=False)
        fig.show()

    def result_chart_stepwise(self):
        fig = make_subplots(specs=[[{'secondary_y': True}]])
        # Add traces
        fig.add_trace(
            go.Scatter(x=self.df[0]['Duration'], y=self.df[0]['Inlet_Pressure'] * 1e-6, name='Messwert Eingang',
                       line_color='blue'))
        fig.add_trace(
            go.Scatter(x=self.df[0]['Duration'], y=self.df[0]['Outlet_Pressure'] * 1e-6, name='Messwert Ausgang',
                       line_color='blue'))
        fig.add_trace(
            go.Scatter(x=self.df[0]['Duration'], y=self.df[0]['Inlet_Pressure_Cal'] * 1e-6, name='Rechenwert Eingang',
                       line_color='red', mode='markers', marker=dict(size=3)))
        fig.add_trace(
            go.Scatter(x=self.df[0]['Duration'], y=self.df[0]['Outlet_Pressure_Cal'] * 1e-6, name='Rechenwert Ausgang',
                       line_color='red', mode='markers', marker=dict(size=3)))

        fig.add_trace(
            go.Scatter(x=self.df[0]['Duration'], y=self.df[0]['Confining_Pressure'] * 1e-6, name='Manteldruck',
                       line_color='black', visible='legendonly'))

        for i in range(1, len(self.df)):
            fig.add_trace(
                go.Scatter(x=[self.df[i].iloc[-1]['Duration']], y=[self.result[i].x[0]], name='k' + str(i), line_color='black',
                           mode='markers', marker=dict(size=10)), secondary_y=True)

        fig.update_layout(
            xaxis=dict(showexponent='all', exponentformat='power'))
        fig.update_xaxes(title_text='Messdauer in s', type='log')
        fig.update_yaxes(title_text='Druck in MPa', secondary_y=False)
        fig.update_yaxes(title_text='Permeabilität in m²',secondary_y=True, showgrid=False)
        fig.show()


class PlotterReaktor(Plotter):

    def raw_data_chart(self):
        fig = make_subplots(specs=[[{'secondary_y': True}]])
        # Add traces
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Inlet_Pressure'] * 1e-6, name='Eingangsdruck',
                       line_color='blue'),
            secondary_y=False)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Outlet_Pressure'] * 1e-6, name='Ausgangsdruck',
                       line_color='blue'),
            secondary_y=False)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Confining_Pressure_Reactor'] * 1e-6, name='Reaktor',
                       line_color='black', visible='legendonly'),
            secondary_y=False)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Confining_Pressure_Sample'] * 1e-6, name='Probe',
                       line_color='black', visible='legendonly'),
            secondary_y=False)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Temperature'] - 273.15, name='Temperatur',
                       line_color='red'),
            secondary_y=True)
        fig.add_vline(
            x=self.df.iloc[self.start]['Duration'], line_dash='dot')
        fig.add_vline(
            x=self.df.iloc[self.stop - 1]['Duration'], line_dash='dot')

        fig.update_layout(
            title=f'<b>{self.name}</b>',
            xaxis=dict(showexponent='all', exponentformat='power'))
        fig.update_xaxes(title_text='Messdauer in s', type='log')
        fig.update_yaxes(title_text='Druck in MPa', secondary_y=False)
        fig.update_yaxes(title_text='Temperatur in °C', secondary_y=True, range=[12, 16], showgrid=False)
        fig.show()

    def result_chart(self):
        self.df['Diff_Inlet'] = (abs(self.df['Inlet_Pressure'] - self.df['Inlet_Pressure_Cal'])
                                 / self.df['Inlet_Pressure'])
        self.df['Diff_Outlet'] = (abs(self.df['Outlet_Pressure'] - self.df['Outlet_Pressure_Cal'])
                                  / self.df['Outlet_Pressure'])

        fig = make_subplots(specs=[[{'secondary_y': True}]])
        # Add traces
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Inlet_Pressure'] * 1e-6, name='Messwert Eingang',
                       line_color='blue'))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Outlet_Pressure'] * 1e-6, name='Messwert Ausgang',
                       line_color='blue'))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Confining_Pressure_Reactor'] * 1e-6, name='Reaktor',
                      line_color='black', visible='legendonly'))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Confining_Pressure_Sample'] * 1e-6, name='Probe',
                      line_color='black', visible='legendonly'))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Temperature'] - 273.15, name='Temperatur',
                       line_color='red', visible='legendonly'),
            secondary_y=True)

        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Inlet_Pressure_Cal'] * 1e-6, name='Rechenwert Eingang',
                       line_color='red', mode='markers', marker=dict(size=3)))
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Outlet_Pressure_Cal'] * 1e-6, name='Rechenwert Ausgang',
                       line_color='red', mode='markers', marker=dict(size=3)))

        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Diff_Inlet'], name='Abweichung Eingang',
                       line_color='green', fill='tozeroy'),
            secondary_y=True)
        fig.add_trace(
            go.Scatter(x=self.df['Duration'], y=self.df['Diff_Outlet'], name='Abweichung Ausgang',
                       line_color='yellow', fill='tozeroy'),
            secondary_y=True)

        fig.update_layout(
            xaxis=dict(showexponent='all', exponentformat='power'))
        fig.update_xaxes(title_text='Messdauer in s', type='log')
        fig.update_yaxes(title_text='Druck in MPa', secondary_y=False)
        fig.update_yaxes(secondary_y=True, showgrid=False)
        fig.show()


import os
import pandas as pd
from .import_data import Data, DataReaktor
from .optimize import Optimizer
from .plots import Plotter, PlotterReaktor


class Measurement:

    def __init__(self, path):
        self.path = path
        self.file_name, _ = os.path.splitext(os.path.split(path)[1])
        self.path_raw, _ = os.path.splitext(os.path.split(path)[0])
        self.path_sim = os.path.join(os.path.dirname(os.path.dirname(path)), 'sim_data')
        self.df_100 = None
        self.df_final = None
        self.sample_data = None

    def calculate_permeability(self, guess, parameter='k'):
        if self.find_file():
            user_input = input('''
            This measurement has already been evaluated. 
            Do you want to use the time adjusted measurement file? (y/n)''')
            if user_input == 'y':
                self.set_adjusted_data()
            elif user_input == 'n':
                self.set_data()
        else:
            self.set_data()

        result = Optimizer(self.df_100, self.sample_data, guess)
        result, opt_steps = result.nelder_mead(parameter)

        plot = Plotter(self.df_100, **{'name': self.file_name})
        plot.result_chart()

        self.add_results(result, guess)
        user_input = input('Do you want to save the results? (y)')
        if user_input == 'y':
            self.save_adjusted_measurement_file()
            self.save_results()

    def calculate_permeability_stepwise(self, guess, parameter='k'):
        self.set_data()
        data = Data(self.path)
        df_100_list = []
        result_list = []
        temp = pd.DataFrame({'t': [], 'k': []})

        result = Optimizer(self.df_100, self.sample_data, guess)
        result, opt_steps = result.nelder_mead(parameter)
        df_100_list.append(self.df_100)
        result_list.append(result)
        temp.loc[0] = [df_100_list[0]['Duration'].max(), result_list[0].x[0]]

        duration_10_percent = self.df_final['Duration'].max() * 0.1
        for i in range(9):
            self.df_100 = self.df_final[self.df_final['Duration'].between(1, duration_10_percent*(i+1))]
            self.df_100.reset_index(inplace=True, drop=True)
            self.df_100 = data.interpolate(self.df_100)
            result = Optimizer(self.df_100, self.sample_data, guess)
            result, opt_steps = result.nelder_mead(parameter)
            df_100_list.append(self.df_100)
            result_list.append(result)
            temp.loc[i+1] = [df_100_list[-1]['Duration'].max(), result_list[-1].x[0]]

        plot = Plotter(df_100_list, **{'name': self.file_name, 'result': result_list})
        plot.result_chart_stepwise()

        df = df_100_list[0][['Duration', 'DateTime', 'Inlet_Pressure', 'Inlet_Pressure_Cal', 'Outlet_Pressure',
                             'Outlet_Pressure_Cal', 'Confining_Pressure', 'Temperature']]

        path = os.path.join(self.path_sim, self.file_name + '_interval.txt')
        df_res = pd.DataFrame.from_dict(self.sample_data, orient='index')
        df_res.to_csv(path, header=False, sep=':', float_format='%.4f')
        df.to_csv(path, index=False, mode='a', float_format='%.2f')
        temp.to_csv(path, mode='a', index=False)

    def get_solution(self):
        df = pd.read_csv(os.path.join(self.path_sim, self.file_name + '.csv'),
                         nrows=10, sep=':', index_col=0, header=None)
        return [float(df.loc['k', 1]), float(df.loc['n', 1])]

    def set_adjusted_data(self):
        data = Data(self.path)
        self.df_100, self.df_final = data.adjusted_pressure_file()
        self.sample_data = data.sample_data()

    def set_data(self):
        data = Data(self.path)
        self.df_100, self.df_final = data.new_pressure_file()
        self.sample_data = data.sample_data()

    def add_results(self, result, guess):
        try:
            porosity = result.x[1]
        except:
            porosity = guess[1]
        self.sample_data.update({'k': result.x[0],
                                 'n': porosity,
                                 'error': result.fun / 100})

    def save_results(self):
        df = self.df_100[['Duration', 'DateTime', 'Inlet_Pressure', 'Inlet_Pressure_Cal', 'Outlet_Pressure',
                          'Outlet_Pressure_Cal', 'Confining_Pressure', 'Temperature']]

        path = os.path.join(self.path_sim, self.file_name + '.csv')
        df_res = pd.DataFrame.from_dict(self.sample_data, orient='index')
        df_res.to_csv(path, header=False, sep=':', float_format='%.4f')
        df.to_csv(path, index=False, mode='a', float_format='%.2f')
        self.df_final.describe().to_csv(path, mode='a', float_format='%.2f')

    def save_adjusted_measurement_file(self):
        path = os.path.join(self.path_raw, self.file_name + '_adjusted.csv')
        self.df_final.to_csv(path, sep=',', index=False, float_format='%.2f')

    def find_file(self):
        for root, dirs, files in os.walk(self.path_raw):
            if self.file_name + '_adjusted.csv' in files:
                return True
            else:
                return False


class MeasurementReaktor(Measurement):

    def calculate_permeability(self, guess, parameter='k'):
        if self.find_file():
            user_input = input('''
            This measurement has already been evaluated. 
            Do you want to use the time adjusted measurement file? (y/n)''')
            if user_input == 'y':
                self.set_adjusted_data()
            elif user_input == 'n':
                self.set_data()
        else:
            self.set_data()
        result = Optimizer(self.df_100, self.sample_data, guess)
        result, opt_steps = result.nelder_mead(parameter)

        plot = PlotterReaktor(self.df_100, **{'name': self.file_name})
        plot.result_chart()

        self.add_results(result, guess)
        user_input = input('Do you want to save the results? (y)')
        if user_input == 'y':
            self.save_adjusted_measurement_file()
            self.save_results()

    def set_data(self):
        data = DataReaktor(self.path)
        self.df_100, self.df_final = data.new_pressure_file()
        self.sample_data = data.sample_data()

    def set_adjusted_data(self):
        data = DataReaktor(self.path)
        self.df_100, self.df_final = data.adjusted_pressure_file()
        self.sample_data = data.sample_data()

    def save_results(self):
        df = self.df_100[['Duration', 'DateTime', 'Inlet_Pressure', 'Inlet_Pressure_Cal', 'Outlet_Pressure',
                          'Outlet_Pressure_Cal', 'Confining_Pressure_Reactor', 'Confining_Pressure_Sample',
                          'Temperature']]

        path = os.path.join(self.path_sim, self.file_name + '.csv')
        df_res = pd.DataFrame.from_dict(self.sample_data, orient='index')
        df_res.to_csv(path, header=False, sep=':', float_format='%.4f')
        df.to_csv(path, index=False, mode='a', float_format='%.2f')
        self.df_final.describe().to_csv(path, mode='a', float_format='%.2f')



import numpy as np
import scipy.sparse
import scipy.sparse.linalg
import CoolProp.CoolProp as cp
from .import_data import Data


class LinearSystem:
    number_of_cells = 50

    def __init__(self, df_100, sample_data, guess):
        self.data = {'inlet_pressure': np.array(df_100['Inlet_Pressure'].values),
                     'outlet_pressure': np.array(df_100['Outlet_Pressure'].values),
                     'duration': np.array(df_100['Duration'].values),
                     'datetime': np.array(df_100['DateTime'].values),
                     'temperature': np.array(df_100['Temperature'].values)}
        self.guess = {'permeability': guess[0],
                      'porosity': guess[1]}
        self.sample = sample_data

    def solve_linear_system(self):
        number_of_timesteps = len(self.data['duration']) - 1
        self.calculate_timesteps()
        self.initialize_calculated_pressure()
        sample_pressure = self.get_initial_pressure()

        for step in range(number_of_timesteps):
            self.data.update({'actual_time_step': step})
            sample_pressure = self.iterate_nonlinear_parameters(sample_pressure)
            self.data['inlet_pressure_calculated'][step+1] = sample_pressure[0]
            self.data['outlet_pressure_calculated'][step+1] = sample_pressure[-1]

        self.data.update({'cell_pressure': sample_pressure})
        return self.data

    def get_initial_pressure(self):
        atmospheric_pressure = self.data['outlet_pressure'][0]
        sample_pressure = np.ones(self.number_of_cells) * atmospheric_pressure
        sample_pressure[0] = self.data['inlet_pressure'][0]
        return sample_pressure

    def initialize_calculated_pressure(self):
        inlet_pressure_calculated = np.ones(Data.number_of_time_steps)
        inlet_pressure_calculated[0] = self.data['inlet_pressure'][0]
        outlet_pressure_calculated = np.ones(Data.number_of_time_steps)
        outlet_pressure_calculated[0] = self.data['outlet_pressure'][0]
        self.data.update({'inlet_pressure_calculated': inlet_pressure_calculated,
                          'outlet_pressure_calculated': outlet_pressure_calculated})

    def calculate_timesteps(self):
        timesteps = np.diff(self.data['duration'])
        self.data.update({'timesteps': timesteps})

    def iterate_nonlinear_parameters(self, sample_pressure):
        i = 0
        difference = 1

        _, solution_vector = self.get_linear_system(sample_pressure)
        while difference > 1e-5 and i <= 10:
            coefficient_matrix, _ = self.get_linear_system(sample_pressure)
            sample_pressure_new = scipy.sparse.linalg.spsolve(coefficient_matrix, solution_vector)
            difference = self.l2_norm(sample_pressure_new, sample_pressure)
            sample_pressure = sample_pressure_new
            i += 1

        return sample_pressure

    @staticmethod
    def residuum(A, x_guess, b):
        r = abs(np.sum((A*x_guess - b) / b))
        return r

    def get_linear_system(self, sample_pressure):
        main_diagonal, off_diagonal, solution_vector = self.build_diagonals(sample_pressure)
        A = scipy.sparse.diags(
            diagonals=[main_diagonal*-1, off_diagonal, off_diagonal],
            offsets=[0, -1, 1],
            shape=(self.number_of_cells, self.number_of_cells),
            format='csr')
        b = - solution_vector * sample_pressure
        return A, b

    def build_diagonals(self, sample_pressure):
        k, n = self.initialize_permeability_porosity()
        compressibility, viscosity, density = self.get_coolprop_data(sample_pressure)
        dx = self.sample['length'] / (self.number_of_cells - 1)
        dt = self.data['timesteps'][self.data['actual_time_step']]
        area = self.sample['area']
        inlet_volume = self.sample['inlet_chamber_volume']
        outlet_volume = self.sample['outlet_chamber_volume']

        viscosity_mean = (viscosity[1:] + viscosity[:-1]) / 2
        density_mean = (density[1:] + density[:-1]) / 2
        k_mean_harmonic = ((2*k[1:]*k[:-1]) / (k[1:]+k[:-1]))

        off_diagonal = (density_mean * area * k_mean_harmonic) / (viscosity_mean * dx)
        main_diagonal = (area * n * compressibility * density * dx) / dt
        main_diagonal[1:-1] = main_diagonal[1:-1] + off_diagonal[:-1] + off_diagonal[1:]
        main_diagonal[0] = (inlet_volume * density[0] * compressibility[0]) / dt + off_diagonal[0]
        main_diagonal[-1] = (outlet_volume * density[-1] * compressibility[-1]) / dt + off_diagonal[-1]
        solution_vector = area * n * compressibility * density * dx / dt
        solution_vector[0] = inlet_volume * density[0] * compressibility[0] / dt
        solution_vector[-1] = outlet_volume * density[-1] * compressibility[-1] / dt

        return main_diagonal, off_diagonal, solution_vector

    def initialize_permeability_porosity(self):
        k = np.ones(self.number_of_cells) * self.guess['permeability']
        k[0] = k[-1] = 1
        n = np.ones(self.number_of_cells) * self.guess['porosity']
        n[0] = n[-1] = 1
        return k, n

    def get_coolprop_data(self, pressure):
        try:
            parameter = ['ISOTHERMAL_COMPRESSIBILITY', 'VISCOSITY', 'DMASS']
            result = cp.PropsSI(parameter, 'T', self.data['temperature'].mean(), 'P', pressure, self.sample['gas'])
            result = result.T
        except:
            result = [np.zeros(self.number_of_cells),np.zeros(self.number_of_cells),np.zeros(self.number_of_cells)]
        return result

    @staticmethod
    def l2_norm(p, p_ref):
        return np.linalg.norm(p-p_ref, 2) / np.linalg.norm(p_ref, 2)




import os.path
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
import datetime as dt
import CoolProp.CoolProp as cp
from .plots import Plotter, PlotterReaktor


class Data:
    number_of_time_steps = 100

    def __init__(self, path):
        self.path = path
        self.file_name, _ = os.path.splitext(os.path.split(path)[1])
        self.path_raw, _ = os.path.splitext(os.path.split(path)[0])
        self.path_base = os.path.split(self.path_raw)[0]
        self.start = None
        self.stop = None

    @staticmethod
    def read_file(path):
        try:
            df = pd.read_csv(path, sep=' ')
            return df
        except Exception as ex:
            print(f'Exception {type(ex).__name__}, {ex.args}')

    @staticmethod
    def convert_units(df):
        df['DateTime'] = pd.to_datetime(df['Date'] + df['Time'], format='%d.%m.%Y%X')
        df['Duration'] = pd.to_timedelta(df['DateTime'] - df['DateTime'][0]).dt.total_seconds()
        df.drop(['Date', 'Time'], axis=1, inplace=True)
        df = df[['Duration', 'DateTime', 'Inlet_Pressure', 'Outlet_Pressure', 'Confining_Pressure', 'Temperature']]
        # convert bar (relative) to Pascal (absolute) and °C to K
        df[['Inlet_Pressure', 'Outlet_Pressure', 'Confining_Pressure']] = \
            df[['Inlet_Pressure', 'Outlet_Pressure', 'Confining_Pressure']].apply(lambda x: x * 1e5 + 97700)
        df['Temperature'] = df['Temperature'] + 273.15
        return df

    def adjust_measurement_interval(self, df):
        self.set_start_stop(df)
        self.set_start_stop_manual(df)
        df = df.iloc[self.start:self.stop]
        print(df.describe())
        df = self.reset_duration(df)
        return df

    def set_start_stop(self, df):
        df['open'] = df['Inlet_Pressure'] < df['Inlet_Pressure'].max() * 0.98
        try:
            open = df.index[df['open'] == True].tolist()[0]
            self.start = df.iloc[open:]['Inlet_Pressure'].idxmax()
        except IndexError:
            self.start = 1

        df['pressure_equilibrium'] = df['Inlet_Pressure'] <= df['Outlet_Pressure']
        try:
            self.stop = df.index[df['pressure_equilibrium'] == True].tolist()[0]
        except:
            self.stop = len(df)
        df.drop(['open', 'pressure_equilibrium'], axis=1, inplace=True)

    def set_start_stop_manual(self, df):
        while True:
            user_input = self.show_plot(df)
            if user_input == 'y':
                break
            else:
                start, stop = user_input.split(',')
                if start.isdigit() & stop.isdigit():
                    self.start = df['Duration'].sub(int(start)).abs().idxmin()
                    self.stop = df['Duration'].sub(int(stop)).abs().idxmin()
                elif start.isdigit():
                    self.start = df['Duration'].sub(int(start)).abs().idxmin()
                elif stop.isdigit():
                    self.stop = df['Duration'].sub(int(stop)).abs().idxmin()

    def show_plot(self, df):
        #df = self.calculate_k_analytic(df)
        plot = Plotter(df, **{'start': self.start, 'stop': self.stop, 'name': self.file_name})
        plot.raw_data_chart()
        user_input = input('''
Measurement interval correctly estimated? (y)
If not enter start and end time manually as integers separated by a comma. (start, end)
        ''')
        return user_input

    def calculate_k_analytic(self, df):
        sample = self.sample_data()
        df['Viscosity'] = cp.PropsSI('VISCOSITY', 'T', df['Temperature'].to_list(),
                                     'P', df['Outlet_Pressure'].to_list(), sample['gas'])
        V = sample['outlet_chamber_volume']
        L = sample['length']
        A = sample['area']

        df['k'] = - ((V * df['Viscosity'] * 2 * L * df['Outlet_Pressure'].diff()) /
                     (A * (df['Outlet_Pressure']**2 - df['Inlet_Pressure']**2) * df['Duration'].diff()))
        return df

    @staticmethod
    def interpolate(df):
        start_date_in_seconds = (df['DateTime'] - dt.datetime(1970,1,1)).dt.total_seconds()[0] - 1
        time_log_scale = np.geomspace(int(df['Duration'].min()), int(df['Duration'].max()), Data.number_of_time_steps).round(2)

        function_inlet = interp1d(df['Duration'], df['Inlet_Pressure'])
        function_outlet = interp1d(df['Duration'], df['Outlet_Pressure'])
        function_confining = interp1d(df['Duration'], df['Confining_Pressure'])
        function_temperature = interp1d(df['Duration'], df['Temperature'])

        inlet_log_scale = function_inlet(time_log_scale)
        outlet_log_scale = function_outlet(time_log_scale)
        confining_log_scale = function_confining(time_log_scale)
        temperature_log_scale = function_temperature(time_log_scale)

        df = pd.DataFrame(np.array([time_log_scale, inlet_log_scale, outlet_log_scale,
                                    confining_log_scale, temperature_log_scale]).T,
                          columns=['Duration', 'Inlet_Pressure', 'Outlet_Pressure',
                                   'Confining_Pressure', 'Temperature'])
        df['DateTime'] = pd.to_datetime(df['Duration']+start_date_in_seconds, unit='s')
        return df

    @staticmethod
    def reset_duration(df):
        df = df.copy()
        df.reset_index(inplace=True, drop=True)
        df['Duration'] = df['Duration'] - df.iloc[0]['Duration'] + 1
        return df

    def new_pressure_file(self):
        df = self.read_file(self.path)
        df = self.drop_duplicates(df)
        df = self.convert_units(df)
        df_final = self.adjust_measurement_interval(df)
        df_final_100 = self.interpolate(df_final)
        return df_final_100, df_final

    def adjusted_pressure_file(self):
        df_final = pd.read_csv(os.path.join(self.path_raw, self.file_name + '_adjusted.csv'), parse_dates=['DateTime'])
        df_final_100 = self.interpolate(df_final)
        return df_final_100, df_final

    def drop_duplicates(self, df):
        rows = len(df)
        df.drop_duplicates(ignore_index=True, inplace=True)
        if rows > len(df):
            df.to_csv(self.path, sep=' ', index=False)
            print(f'Dropped {rows-len(df)} duplicate rows.')
        return df

    def sample_data(self):
        my_dict = self.get_core_dimensions()
        unit = self.get_unit_dimensions()
        my_dict.update({'inlet_chamber_volume': unit[0],
                        'outlet_chamber_volume': unit[1]})
        return my_dict

    def get_core_dimensions(self):
        database = self.read_file(os.path.join(self.path_base, 'database.csv'))
        database.set_index('name', inplace=True)

        length = database.loc[self.file_name, 'length']
        diameter = database.loc[self.file_name, 'diameter']
        area = np.pi * 0.25 * diameter**2
        gas = database.loc[self.file_name, 'gas']
        my_dict = {'length': length,
                   'diameter': diameter,
                   'area': area,
                   'gas': gas}
        return my_dict

    def get_unit_dimensions(self):
        database = self.read_file(os.path.join(self.path_base, 'database.csv'))
        units = self.read_file(os.path.join(self.path_base, 'measurement_units.csv'))
        ml_to_m3 = 1e-6

        database.set_index('name', inplace=True)
        unit = database.loc[self.file_name, 'unit']
        filt = units['number'] == unit
        volume = units.loc[filt, ['inlet_chamber_in_ml', 'outlet_chamber_in_ml']]
        return volume.values[0] * ml_to_m3


class DataReaktor(Data):

    @staticmethod
    def convert_units(df):
        df['DateTime'] = pd.to_datetime(df['Date'] + df['Time'], format='%d.%m.%Y%X')
        df['Duration'] = pd.to_timedelta(df['DateTime'] - df['DateTime'][0]).dt.total_seconds()
        df.drop(['Date', 'Time'], axis=1, inplace=True)
        df = df[['Duration', 'DateTime', 'Inlet_Pressure', 'Outlet_Pressure', 'Confining_Pressure_Reactor',
                 'Confining_Pressure_Sample', 'Temperature']]
        # convert bar (relative) to Pascal (absolute) and °C to K
        df[['Inlet_Pressure', 'Outlet_Pressure', 'Confining_Pressure_Reactor', 'Confining_Pressure_Sample']] = \
            df[['Inlet_Pressure', 'Outlet_Pressure', 'Confining_Pressure_Reactor', 'Confining_Pressure_Sample']]\
                .apply(lambda x: x * 1e5 + 97700)
        df['Temperature'] = df['Temperature'] + 273.15
        return df

    def show_plot(self, df):
        plot = PlotterReaktor(df, **{'start': self.start, 'stop': self.stop, 'name': self.file_name})
        plot.raw_data_chart()
        user_input = input('''
        Measurement interval correctly estimated? (y)
        If not enter start and end time manually as integers separated by a comma. (start, end)
                ''')
        return user_input

    @staticmethod
    def interpolate(df):
        start_date_in_seconds = (df['DateTime'] - dt.datetime(1970,1,1)).dt.total_seconds()[0] - 1
        time_log_scale = np.geomspace(1, int(df['Duration'].max()), 100).round(2)

        function_inlet = interp1d(df['Duration'], df['Inlet_Pressure'])
        function_outlet = interp1d(df['Duration'], df['Outlet_Pressure'])
        function_confining_reaktor = interp1d(df['Duration'], df['Confining_Pressure_Reactor'])
        function_confining_sample = interp1d(df['Duration'], df['Confining_Pressure_Sample'])
        function_temperature = interp1d(df['Duration'], df['Temperature'])

        inlet_log_scale = function_inlet(time_log_scale)
        outlet_log_scale = function_outlet(time_log_scale)
        reaktor_log_scale = function_confining_reaktor(time_log_scale)
        sample_log_scale = function_confining_sample(time_log_scale)
        temperature_log_scale = function_temperature(time_log_scale)

        df = pd.DataFrame(np.array([time_log_scale, inlet_log_scale, outlet_log_scale,
                                    reaktor_log_scale, sample_log_scale, temperature_log_scale]).T,
                          columns=['Duration', 'Inlet_Pressure', 'Outlet_Pressure',
                                   'Confining_Pressure_Reactor', 'Confining_Pressure_Sample', 'Temperature'])
        df['DateTime'] = pd.to_datetime(df['Duration']+start_date_in_seconds, unit='s')
        return df

    def adjusted_pressure_file(self):
        df_final = pd.read_csv(os.path.join(self.path_raw, self.file_name + '_adjusted.csv'), parse_dates=['DateTime'])
        df_final_100 = self.interpolate(df_final)
        return df_final_100, df_final

    def get_core_dimensions(self):
        database = self.read_file(os.path.join(self.path_base, 'database_reaktor.csv'))
        database.set_index('name', inplace=True)

        length = database.loc[self.file_name, 'length']
        outer_diameter = database.loc[self.file_name, 'outer_diameter']
        inner_diameter = database.loc[self.file_name, 'inner_diameter']
        area = np.pi * 0.25 * (outer_diameter**2 - inner_diameter**2)
        gas = database.loc[self.file_name, 'gas']
        my_dict = {'length': length,
                   'outer_diameter': outer_diameter,
                   'inner_diameter': inner_diameter,
                   'area': area,
                   'gas': gas}
        return my_dict

    def get_unit_dimensions(self):
        database = self.read_file(os.path.join(self.path_base, 'database_reaktor.csv'))
        units = self.read_file(os.path.join(self.path_base, 'measurement_units.csv'))
        ml_to_m3 = 1e-6

        database.set_index('name', inplace=True)
        unit = database.loc[self.file_name, 'unit']
        filt = units['number'] == unit
        volume = units.loc[filt, ['inlet_chamber_in_ml', 'outlet_chamber_in_ml']]
        return volume.values[0] * ml_to_m3






import numpy as np
import pandas as pd
import scipy.optimize as optimize
from .linear_system import LinearSystem


class Optimizer:

    def __init__(self, df_100, sample_data, guess):
        self.df_100 = df_100
        self.sample_data = sample_data
        self.guess = guess
        self.data = None
        self.optimization_steps = [['k', 'n', 'e']]

    def nelder_mead(self, parameter):
        if parameter == 'k':
            min_result = optimize.minimize(self.optimize_function, self.guess[0], args=parameter,
                                           method='Nelder-Mead', options={'disp': False}, tol=0.001)
        elif parameter == 'both':
            min_result = optimize.minimize(self.optimize_function, self.guess, args=parameter,
                                           method='Nelder-Mead', options={'disp': False}, tol=0.001)

        self.print_final_result(min_result)

        self.df_100[['Inlet_Pressure_Cal', 'Outlet_Pressure_Cal']] = None
        self.df_100['Inlet_Pressure_Cal'] = pd.DataFrame.from_dict(self.data['inlet_pressure_calculated'])
        self.df_100['Outlet_Pressure_Cal'] = pd.DataFrame.from_dict(self.data['outlet_pressure_calculated'])

        return min_result, self.optimization_steps

    def optimize_function(self, guess, parameter):
        if parameter == 'k':
            guess = [guess[0], self.guess[1]]
        elif parameter == 'both':
            guess = guess

        self.data = LinearSystem(self.df_100, self.sample_data, guess).solve_linear_system()
        error = self.calculate_error()
        self.optimization_steps.append([guess[0], guess[1], error/100])
        print(f'k = {guess[0]:.4} m^2, n = {guess[1]:.4}, e = {error:.3} %')
        return error

    def calculate_error(self):
        p_in_ref = self.data['inlet_pressure']
        p_in = self.data['inlet_pressure_calculated']
        p_out_ref = self.data['outlet_pressure']
        p_out = self.data['outlet_pressure_calculated']

        absolute_magnitude = np.sqrt(sum(p_in_ref**2 + p_out_ref**2))
        difference = abs(p_in_ref - p_in) + abs(p_out_ref - p_out)
        absolute_error = np.sqrt(sum(difference**2))
        relative_error = absolute_error / absolute_magnitude * 100
        return relative_error

    @staticmethod
    def print_final_result(min_result):
        print(f'\n Calculation finished: {min_result.message} \n'
              f'\tNumber of iterations: {min_result.nit} \n'
              f'\tNumber of function evaluations: {min_result.nfev} \n'
              f'\tPermeability: {min_result.x[0]:.6} m^2 \n'
              f'\tPorosity: \n'
              f'\tRelative error: {round(min_result.fun, 2)} %')
